from ...kernel.system.Adm import Adm
from ...kernel.agent.Action import Action

class RESTRequestAction(Action):
	
    def execute(self, data):
        url = data['url']
        dto = data['data']

        self.agent.requestID = self.agent.requestID + 1
        rqstID = str(self.agent.requestID)
        self.agent.requestTable[rqstID] = data['queue'] 

        dto['requestID'] = rqstID
        urlSplit = url.split('/')
        Adm().sendEvent(urlSplit[0], urlSplit[1], dto)




# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
------------------------- PBESA --------------------------
----------------------------------------------------------

@autor AKEN & SIDRE
@version 3.0.1
@date 27/07/20
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------
import time
from pbesa.kernel.system.Adm import Adm
from pbesa.social.worker.task import Task
from pbesa.kernel.agent.Agent import Agent
from pbesa.kernel.agent.Action import Action
from pbesa.social.worker.worker import Worker
from pbesa.social.linealcontroller.delegateaction import DelegateAction
from pbesa.social.linealcontroller.responseaction import ResponseAction
from pbesa.social.linealcontroller.linealcontroller import LinealController

# --------------------------------------------------------
# Define controller agent
# --------------------------------------------------------

# --------------------------------------------------------
# Define Action
class TranslateDelegate(DelegateAction):
    """ An action is a response to the occurrence of an event """

    def delegate(self, data):
        """
        Catch the exception.
        @param exception Response exception
        """
        #
        tokens = data.split('_')
        self.toAssign(tokens[0])
        self.toAssign(tokens[1])
        self.activeTimeout(3)

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        pass

# --------------------------------------------------------
# Define Action
class TranslateResponse(ResponseAction):
    """ An action is a response to the occurrence of an event """

    def endOfProcess(self, resultDict, timeout):
        """
        Catch the exception.
        @param exception Response exception
        """
        result = "%s %s" % (resultDict['w1'], resultDict['w2'])
        self.sendResponse(result)

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        pass

# --------------------------------------------------------
# Define Agent
class TranslateController(LinealController):
    """ Through a class the concept of agent is defined """
    
    def build(self):
        """
        Method that allows defining the structure and 
        resources of the agent
        """
        # Assign an action to the behavior
        self.bindDelegateAction(TranslateDelegate())
        # Assign an action to the behavior
        self.bindResponseAction(TranslateResponse())

    def shutdown(self):
        """ Method to free up the resources taken by the agent """
        pass

# --------------------------------------------------------
# Define worker agent
# --------------------------------------------------------

# --------------------------------------------------------
# Define Action
class TranslateTask(Task):
    """ An action is a response to the occurrence of an event """

    def goHead(self, data):
        """
        Execute.
        @param data Event data
        """
        if data == 'Hello':
            self.sendResponse('Hola')
        if data == 'World':
            self.sendResponse('Mundo')

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        pass

# --------------------------------------------------------
# Define Agent
class WorkerAgent(Worker):
    """ Through a class the concept of agent is defined """
    
    def build(self):
        """
        Method that allows defining the structure and 
        resources of the agent
        """
        # Assign an action to the behavior
        self.bindTask(TranslateTask())
        
    def shutdown(self):
        """ Method to free up the resources taken by the agent """
        pass

# --------------------------------------------------------
# Main
# --------------------------------------------------------
if __name__ == "__main__":
    """ Main """
    # Initialize the container
    mas = Adm()
    mas.start()

    #-----------------------------------------------------
    # Create the worker agents
    w1ID = 'w1'
    w1 = WorkerAgent(w1ID)
    w1.start()
    w2ID = 'w2'
    w2 = WorkerAgent(w2ID)
    w2.start()
    
    #-----------------------------------------------------
    # Create the controller agent
    ctrID = 'Jarvis'
    ag = TranslateController(ctrID)
    ag.suscribeAgent(w1)
    ag.suscribeAgent(w2)
    ag.start()

    # Call
    result = mas.callAgent(ctrID, "Hello_World")

    # Log
    print(result)



from ....kernel.agent.Action import Action

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------
class RetractFactAction(Action):
    
    def execute(self, data):
        self.agent.engine.retract_fact(data)

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        pass




# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
------------------------- PBESA --------------------------
----------------------------------------------------------

@autor AKEN & SIDRE
@version 3.0.1
@date 27/07/20
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------
import time
import traceback
from pbesa.kernel.system.Adm import Adm
from pbesa.kernel.agent.Agent import Agent
from pbesa.kernel.agent.Action import Action

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------
class SumAction(Action):
    """ An action is a response to the occurrence of an event """

    def execute(self, data):
        """ 
        Response.
        @param data Event data 
        """
        print(self.agent.state['acum'] + data)

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        pass

# --------------------------------------------------------
# Define Agent
# --------------------------------------------------------
class SumAgent(Agent):
    """ Through a class the concept of agent is defined """
    
    def setUp(self):
        """
        Method that allows defining the status, structure 
        and resources of the agent
        """
        # Defines the agent state
        self.state = {
            'acum': 7
        }
        # Defines the behavior of the agent. An agent can 
        # have one or many behaviors
        self.addBehavior('calculate')
        # Assign an action to the behavior
        self.bindAction('calculate', 'sum', SumAction())

    def shutdown(self):
        """ Method to free up the resources taken by the agent """
        pass

# --------------------------------------------------------
# Main
# --------------------------------------------------------
if __name__ == "__main__":
    """ Main """
    try:
        # Initialize the container
        mas = Adm()
        mas.start()

        # Create the agent
        agentID = 'Jarvis'
        ag = SumAgent(agentID)
        ag.start()

        # Send the event
        data = 8
        mas.sendEvent('Jarvis', 'sum', data)

        # Remove the agent from the system
        time.sleep(1)
        mas.killAgent(ag)

        # Destroy the Agent Container
        mas.destroy()
    except:
        traceback.print_exc()



from abc import abstractmethod
from ...kernel.agent.Agent import Agent
from .exceptions import DelegatorException

class Delegator(Agent):
    
    def __init__(self, agentID):
        self.__agentList = []
        super().__init__(agentID)

    def setUp(self):
        self._social = True
        # Defines the agent state
        self.state = {}
        # Defines the delegate behavior        
        self.addBehavior('Delegate')
        # Defines the space for build
        self.build()

    def bindDelegateAction(self, action):
        self.bindAction('Delegate', 'delegate', action)

    def suscribeAgent(self, agent):
        if not isinstance(agent, Agent):
            raise DelegatorException('[Warn, suscribeAgent]: The object to subscribe is not an agent')
        self.__agentList.append(agent.id)

    @abstractmethod
    def build(self):
        pass

    def start(self):
        if len(self.__agentList) <= 0:
            raise DelegatorException('[Warn, start]: Delegate agent list is empty. Agents must be subscribed to the Delegate Agent')
        super().start()

    def getAgentList(self):
        return self.__agentList


from .psagent.rule import FunctionConsequent, Parameter, TestAntecedent

class BethaNode():

    def __init__(self):
        self.name = None
        self.fire = False
        self.__consequents = []
        self.__alpha_node_dict = {}
        self.__antecedent_list = []

    def add_alpha_node(self, alpha_node):
        self.__alpha_node_dict[alpha_node.pattern] = alpha_node

    def add_consequent(self, consequent):
        self.__consequents.append(consequent)

    def add_antecedent(self, antecedent):
        self.__antecedent_list.append(antecedent)

    def add_element_list(self, lst, element):
        no_exist = True
        for ele in lst:
            if ele.id == element.id:
                no_exist = False
        if no_exist:
            lst.append(element)

    def evaluate(self):
        #print("")
        #print("---------------------------------[%s]:" % self.name)
        var_dict = {}
        test_list = []
        tuple_mlist = []
        template_dict = {}
        #-------------------------------------------------
        # Agrupa los templates
        for antecedent in self.__antecedent_list:
            if isinstance(antecedent, TestAntecedent):
                test_list.append(antecedent)
            else:
                if not antecedent.template in template_dict:
                    template_dict[antecedent.template] = []
                template_dict[antecedent.template].append(antecedent)
        #-------------------------------------------------
        # Compardores de valor
        # Itera los templates
        memory_mlist = []
        for template, antecedent_list in template_dict.items():
            #print("template", template)
            alpha_node = self.__alpha_node_dict[template]
            memory = alpha_node.alpha_memory
            # Filtra la memoria iterando y evaluando los 
            # antecedentes
            for antecedent in antecedent_list:
                fact_list = []
                for fact in memory:
                    if not antecedent.value and antecedent.template_property in fact.obj:
                        # Caso que el valor del 
                        # antecedente es nulo
                        self.add_element_list(fact_list, fact)
                    elif isinstance(antecedent.value, str) and "?" in antecedent.value:
                        # Caso en el que el valor 
                        # es una varible
                        self.add_element_list(fact_list, fact)
                        # Agrupa las variables
                        if not antecedent.value in var_dict:
                            var_dict[antecedent.value] = []
                        var_dict[antecedent.value].append(antecedent) 
                    else:
                        # Caso que el antecedente refiere 
                        # a un objeto
                        property_value = fact.obj[antecedent.template_property]
                        if antecedent.template_property in fact.obj: 
                            if property_value == antecedent.value:
                                # Caso en que el valor 
                                # coincide
                                self.add_element_list(fact_list, fact)
                memory = fact_list
            memory_mlist.append(memory)
        #-----------------------------------------
        # Genera la combinacion
        for fact_list in memory_mlist:
            list_b = []
            result_list = []
            for fact_a in fact_list:
                result_list.append([fact_a])
            list_a = result_list
            if len(tuple_mlist) > 0:
                list_b = tuple_mlist
                result_list = []
                for fact_a in list_a:
                    for fact_b in list_b:
                        com_tmp = []
                        com_tmp.extend(fact_a)
                        com_tmp.extend(fact_b)
                        result_list.append(com_tmp)
                tuple_mlist = result_list
            else:
                tuple_mlist = list_a
        #print("===== Comparadores de Valor ======")
        #print(len(tuple_mlist))
        """
        for fct_list in tuple_mlist:
            for fact in fct_list:
                print("[fact>", fact.template, "]:", fact.__dict__['id'][-10:])
            print("----------------------------------")
        """
        #print("==================================")
        # Valida Cardinalidad
        if len(tuple_mlist) > 0:
            #print("template_dict", len(template_dict))
            #print("tuple_mlist[0]", len(tuple_mlist[0]))
            if not len(template_dict) == len(tuple_mlist[0]):
                tuple_mlist = []

        #-------------------------------------------------
        # Valida los comparadores de operacion
        if len(test_list) and len(tuple_mlist) > 0:
            index = 0
            rm_mlist = []
            # Itera tupla por tupla
            for fct_list in tuple_mlist:
                match = True
                # Itera los comparadores de operacion
                # binarios
                for antecedent in test_list:
                    #-------------------------------------
                    # Obtiene el primer parametro
                    fact_param1 = {}
                    param1 = antecedent.parameters[0]
                    if isinstance(param1, Parameter):
                        # Caso variable  
                        for i in range(0, len(fct_list)):
                            fact = fct_list[i]
                            if fact.template == param1.template:
                                fact_param1['index'] = i
                                fact_param1['obj'] = fact
                                fact_param1['value'] = fact.obj[param1.template_property]
                    else:
                        # Caso valor
                        fact_param1['index'] = None
                        fact_param1['obj'] = None
                        fact_param1['value'] = param1
                    #-------------------------------------
                    # Obtiene el segundo parametro
                    fact_param2 = {}
                    param2 = antecedent.parameters[1]
                    if isinstance(param2, Parameter):
                        # Caso variable    
                        for i in range(0, len(fct_list)):
                            fact = fct_list[i]
                            if fact.template == param2.template:
                                fact_param2['index'] = i
                                fact_param2['obj'] = fact
                                fact_param2['value'] = fact.obj[param2.template_property]
                    else:
                        # Caso valor
                        fact_param2['obj'] = None
                        fact_param2['obj'] = None
                        fact_param2['value'] = param2
                    #-------------------------------------
                    # Evalua la operacion
                    if fact_param1 and fact_param2:                            
                        if antecedent.operation == "=":
                            match &= fact_param1['value'] == fact_param2['value']
                        elif antecedent.operation == ">":
                            match &= fact_param1['value'] > fact_param2['value']
                        elif antecedent.operation == "<":
                            match &= fact_param1['value'] < fact_param2['value']
                        elif antecedent.operation == ">=":
                            match &= fact_param1['value'] >= fact_param2['value']
                        elif antecedent.operation == "<=":
                            match &= fact_param1['value'] <= fact_param2['value']
                    else:
                        # Caso que el template no coincide
                        match = False
                if not match:
                    # Caso donde no pasa la operacion se 
                    # mantiene el indice para escluir 
                    # la tupla
                    rm_mlist.append(index)
                index += 1
            #---------------------------------------------
            # Elimina las tuplas que no pasaron
            if len(rm_mlist) > 0:
                index = 0
                new_tuple_mlist = []
                for fct_list in tuple_mlist:
                    if not index in rm_mlist:
                        new_tuple_mlist.append(fct_list)
                    index += 1
                tuple_mlist = new_tuple_mlist 
            #print("=== Comparadores de Operacion ====")
            #print(len(tuple_mlist))
            """
            print("=> rm_mlist", len(rm_mlist))
            for fct_list in tuple_mlist:
                for fact in fct_list:
                    print("[fact>", fact.template, "]:", fact.__dict__['id'][-10:])
                print("----------------------------------")
            """
            #print("==================================")
            
        #-------------------------------------------------
        # Valida los comparadores de relacion
        if len(var_dict) and len(tuple_mlist) > 0:
            index = 0
            rm_mlist = []
            # Itera tupla por tupla
            for fct_list in tuple_mlist:
                #print("+++", index)
                # Itera los grupos de variable
                for key, antecedent_list in var_dict.items():
                    #print("---", key)
                    value = None
                    # Itera los antecedentes del grupo
                    for antecedent in antecedent_list:
                        # Valida los hechos de la tupla 
                        # reacionados
                        for fact in fct_list:
                            if fact.template == antecedent.template:
                                if not value:
                                    value = fact.obj[antecedent.template_property]
                                """
                                print(
                                    "t:", antecedent.template, 
                                    "p:", antecedent.template_property,
                                    "vp:", fact.obj[antecedent.template_property][-10:],
                                    "v;", value[-10:])
                                """
                                if not value == fact.obj[antecedent.template_property]:
                                    if not index in rm_mlist:
                                        rm_mlist.append(index)
                                        #print(rm_mlist)
                index += 1
            #---------------------------------------------
            # Elimina las tuplas que no pasaron
            if len(rm_mlist) > 0:
                index = 0
                new_tuple_mlist = []
                for fct_list in tuple_mlist:
                    if not index in rm_mlist:
                        new_tuple_mlist.append(fct_list)
                    index += 1
                tuple_mlist = new_tuple_mlist
            #print("==== Comparadores de Relacion ====")
            #print(len(tuple_mlist))
            """
            print("=> rm_mlist", len(rm_mlist))
            for fct_list in tuple_mlist:
                for fact in fct_list:
                    print("[fact>", fact.template, "]:", fact.__dict__['id'][-10:])
                print("----------------------------------")
            """
            #print("==================================")

        #-------------------------------------------------
        # Verifica y dispara
        self.fire = False
        if len(tuple_mlist) > 0:
            if len(tuple_mlist) > 1:
                # Caso de mas de una tupla
                for fct_list in tuple_mlist:
                    #print("")
                    #print("________________________")
                    for consequent in self.__consequents:
                        if isinstance(consequent, FunctionConsequent):
                            if len(consequent.parameters) > 0:
                                data = {}
                                for param in consequent.parameters:
                                    #print("param:", param.template, param.template_property)
                                    for fact in fct_list:
                                        if fact.template == param.template:
                                            data[param.template_property] = fact.obj[param.template_property]
                                if data:
                                    consequent.fun(data)
                                    self.fire = True
                            else:
                                consequent.fun()
                                self.fire = True
            elif len(tuple_mlist) > 0:
                # Caso de una sola tupla
                #print("Caso de una solo tupla")
                #print("")
                #print("________________________")
                for consequent in self.__consequents:
                    if isinstance(consequent, FunctionConsequent):
                        if len(consequent.parameters) > 0:
                            data = {}
                            for param in consequent.parameters:
                                #print("param:", param.template, param.template_property)
                                for fact in tuple_mlist[0]:
                                    if fact.template == param.template:
                                        data[param.template_property] = fact.obj[param.template_property]
                            consequent.fun(data)
                            self.fire = True
                        else:
                            consequent.fun()
                            self.fire = True
        if self.fire:
            print("---------------------------------[%s]:" % self.name)
            print(len(tuple_mlist))
            for fct_list in tuple_mlist:
                for fact in fct_list:
                    print("[fact>", fact.template, "]:", fact.__dict__['id'][-10:])
                print("..............................")
            print("---------------------------------")
            print("")



from ...kernel.agent.Action import Action

class WSSNJDelegateAction(Action):
    
    def execute(self, data):
        request = data['request']
        try:             
            dto = {'context': request['context'], 'data': request['data']}

            if request['context'] == 'register' or request['context'] == 'login':
                agID = self.agent.getFreePool()            
                if agID:
                    socket = data['socket']
                    self.adm.setEndPoint(agID, socket)
                    data = request['data']
                    self.agent.suscribeAgent(data['id'], agID)
                    rst = self.adm.sendEvent(agID, request['event'], dto)
                    if not rst:
                        self.responseError(data)
                else:
                    print('WARN::: All agent is busy.')
                    self.responseError(data)
            else:
                agID = None
                if request['context'] == 'logout':
                    data = request['data']
                    agID = self.agent.getAgentID(data['id'])
                    self.agent.toPool(data['id'], agID)
                else:
                    agID = self.agent.getAgentID(request['agent'])                    
                
                rst = self.adm.sendEvent(agID, request['event'], dto)
                if not rst:
                    agID = self.agent.getFreePool()
                    socket = data['socket']
                    self.adm.setEndPoint(agID, socket)
                    self.agent.suscribeAgent(request['agent'], agID)
                    rst = self.adm.sendEvent(agID, request['event'], dto)
                    if not rst:
                        self.responseError(data)

        except Exception as inst:
            print('Controled ecxeption in WSSNJDelegateAction: ')
            print('MSG: ', inst)
            self.responseError(data)
            

    def responseError(self, data):
        dto = {
            'state':'ERROR',
            'context':'None',
            'command': 'None',
            'msg': 'None',
            'options': []
        }
        socket = data['socket']
        dto = str(dto)
        dto = dto.replace("'", "\"") + '\n'
        socket.sendall(dto.encode())            
        


from scrapy.spiders import Rule
from scrapy.linkextractors import LinkExtractor

class CrawlData():

    def __init__(
        self, 
        urls = None, 
        allowed_domains = None, 
        rules = None,
        user_agent_list = None
    ):
        self.__urls = None
        self.__rules = None
        self.__user_agent_list = None
        self.__allowed_domains = None
        if not urls:
            self.__urls = []
        if not rules:
            self.__rules = ()
        if not allowed_domains:
            self.__allowed_domains = []

    def addRule(self, rule):
        self.__rules += (rule,)

    def addLinkExtractorRule(self, pattern):
        self.__rules += (Rule(LinkExtractor(restrict_text=(pattern, )), callback='execute_parse'),)
    
    def getRules(self):
        return self.__rules

    def setUrls(self, urls):
        self.__urls = urls

    def addUrl(self, url):
        self.__urls.append(url)

    def getUrls(self):
        return self.__urls

    def setAllowedDomains(self, allowed_domains):
        self.__allowed_domains = allowed_domains

    def addAllowedDomain(self, allowed_domain):
        self.__allowed_domains.append(allowed_domain)

    def getAllowedDomains(self):
        return self.__allowed_domains

    def setUserAgentList(self, user_agent_list):
        self.__user_agent_list = user_agent_list

    def getUserAgentList(self):
        return self.__user_agent_list



import uuid
import scrapy
import random
import requests
from abc import abstractmethod
from .SpiderData import SpiderData
import scrapy.crawler as crawler
from twisted.internet import reactor
from .exceptions import SpiderException
from ..kernel.agent.Action import Action
from multiprocessing import Process, Queue
from scrapy.utils.log import configure_logging
from scrapy.utils.project import get_project_settings


class SpiderAction(Action, scrapy.Spider):
    
    def __init__(self, cus_urls = "", agent = None):
        self.urls = cus_urls
        self.name = str(uuid.uuid1())
        super().__init__()
        if agent:
            self.agent = agent
            self.log = agent.log
    
    def start_requests(self):
        """Makes the request urls"""
        for url in self.urls:
            yield scrapy.Request(url, callback = self.execute_parse)

    @abstractmethod
    def execute_parse(self, response):
        pass

    def execute(self, data):
        """ 
        Response.
        @param data Event data 
        """
        try:
            if isinstance(data, SpiderData):
                self.agent.state['data'] = data
                self.run_spider(self.agent, data)
            else:
                self.log.error("The object data must be instance of SpiderData")
                raise SpiderException('[Error, execute]: The object data must be instance of SpiderData')
        except Exception as e:
            self.log.error(str(e))
            raise SpiderException('[Warning, execute]: %s' % str(e))
        
    def run_spider(self, agent, data):
        """Run the spider asocited to agent.
        @param spider:class Calss of SpiderAction.
        @param data:SpiderData dato to scrape
        """
        def f(q):
            try:
                s = get_project_settings()
                user_agent_list = data.getUserAgentList()
                user_agent = None
                if len(user_agent_list) > 0:
                    user_agent = random.choice(user_agent_list)
                if user_agent:
                    s.update({
                        "LOG_ENABLED": "True",
                        "TELNETCONSOLE_ENABLED": "False",
                        "USER_AGENT": user_agent
                    })
                else:
                    s.update({
                        "LOG_ENABLED": "True",
                        "TELNETCONSOLE_ENABLED": "False"
                    })
                runner = crawler.CrawlerRunner(s)
                configure_logging({'LOG_FORMAT': '%(levelname)s: %(message)s'})
                deferred = runner.crawl(
                    agent.class_element, 
                    cus_urls = data.getUrls(),
                    agent = agent
                )
                deferred.addBoth(lambda _: reactor.stop())
                reactor.run()
                q.put(None)
            except Exception as e:
                q.put(e)
                self.log.error(str(e))
                raise SpiderException('[Warning, execute]: %s' % str(e))
        try:
            q = Queue()
            p = Process(target=f, args=(q,))
            p.start()
            try:
                self.log.info("Inicia la arana %s en el proceso: %s" % (agent.id, str(p)))
                q.get(timeout=120)
                self.log.info("Finaliza la arana %s en el proceso: %s" % (agent.id, str(p)))
            except:
                pass        
            p.kill()
        except Exception as e:
            self.log.error(str(e))
            raise SpiderException('[Warning, execute]: %s' % str(e))

    def send_extern_event(self, gateway, command, data):
        try:
            requests.post(gateway, json = {
                "command": command,
                "dto": data
            })
        except Exception as e:
            self.log.error(str(e))
            raise SpiderException('[Warning, execute]: %s' % str(e))



from ...kernel.agent.Agent import Agent
from ...middleware.web.behavior.WSRequestAction import WSRequestAction
from ...middleware.web.behavior.WSResponseAction import WSResponseAction
from ...middleware.web.behavior.RESTRequestAction import RESTRequestAction
from ...middleware.web.behavior.RESTResponseAction import RESTResponseAction
from ...middleware.web.behavior.WSSNJDelegateAction import WSSNJDelegateAction

class FrontControllerAgent(Agent):

    poolList = None
    requestID = None
    agentTable = None
    requestTable = None

    def __init__(self):
        self.poolList = []
        self.requestID = 0
        self.agentTable = {}
        self.requestTable = {}
        super().__init__()

    def setUp(self):
        settings = {
            'id': 'front_controller',
            'state': {}, 
            'behaviors': [
                {'name': 'Delegate', 'events':[
                    {'performative': 'rest_request', 'action' : RESTRequestAction()},
                    {'performative': 'rest_response', 'action' : RESTResponseAction()},
                    {'performative': 'ws_request', 'action' : WSRequestAction()},
                    {'performative': 'ws_response', 'action' : WSResponseAction()},
                    {'performative': 'wsnj_delegate', 'action' : WSSNJDelegateAction()}
                ]}
            ]
        }
        return settings

    def getFreePool(self):
        if self.poolList:
            return self.poolList.pop()
        else:
            return None

    def addPool(self, ag):
        self.poolList.append(ag)

    def getAgentID(self, userID):
        if self.agentTable:
            return self.agentTable[str(userID)]
        return None

    def suscribeAgent(self, userID, agID):
        self.agentTable[userID] = agID

    def toPool(self, userID, agID):
        self.agentTable.remove(userID)
        self.addPool(agID)

                        



from ...kernel.system.Adm import Adm
from .psagent.template import TypeSlot

class AlphaNode():
    
    def __init__(self, pattern):
        self.id = None
        self.__children = []
        self.alpha_memory = []
        self.counter_list = []
        self.pattern = pattern.name
        for slot in pattern.slots:
            if slot.slot_type == TypeSlot.KEY:
                self.id = slot.name
                break
        for slot in pattern.slots:
            if slot.slot_type == TypeSlot.COUNTER:
                self.counter_list.append(slot.name)

    def add_beta_node(self, beta_node):
        no_exist = True
        for beta in self.__children:
            if beta.name == beta_node.name:
                no_exist = False
                break
        if no_exist:
            self.__children.append(beta_node)

    def retract(self, fact):
        # Actualiza el ID del fact
        if not fact.id:
            fact.update_id(self.id)
        new_alpha_memory = []
        for fct in self.alpha_memory:
            if not fct.id == fact.id:
                new_alpha_memory.append(fct)
        self.alpha_memory = new_alpha_memory
        db = Adm().getDBConnection()
        db["work_memory"].delete_one({"id": fact.id})
        
    def evaluate(self, fact, propagate_flag):
        if fact.template == self.pattern:
            # Actualiza el ID del fact
            if not fact.id:
                fact.update_id(self.id)
            # Actualiza el objeto
            updatedList = []
            for i in range(0, len(self.alpha_memory)):
                cur_fact = self.alpha_memory[i]
                if cur_fact.id == fact.id:
                    for key, item in fact.obj.items():
                        if key in self.counter_list:
                            cur_fact.obj[key] = cur_fact.obj[key] + item
                        else:
                            cur_fact.obj[key] = item
                    updatedList.append((i, cur_fact))
                    break
            # Actualiza la memoria
            db = Adm().getDBConnection()
            if len(updatedList) > 0:
                for up in updatedList:
                    self.alpha_memory.pop(up[0])
                    self.alpha_memory.append(up[1])
                    db["work_memory"].delete_one({"id": up[1].id})
                    db["work_memory"].insert_one(up[1].to_dict())    
            else:
                self.alpha_memory.append(fact)
                db["work_memory"].insert_one(fact.to_dict())
            # Propaga
            if propagate_flag:
                for child in self.__children:
                    #if not child.fire:
                    #print("=========================>",self.pattern, child.name)
                    child.evaluate()

    def load_fact(self, fact):
        self.alpha_memory.append(fact)



from ...kernel.system.Adm import Adm
from ...kernel.agent.Action import Action

class WSRequestAction(Action):
    
    def execute(self, data):
        request = data['request']
        self.agent.requestID = self.agent.requestID + 1
        rqstID = str(self.agent.requestID)
        self.agent.requestTable[rqstID] = data['socket']
        try:             
            dto = {'requestID': rqstID, 'context': request['context'], 'data': request['data']}
            Adm().sendEvent(request['agent'], request['event'], dto)	    
        except Exception as inst:
            print('Controled ecxeption: ')
            print('MSG: ', inst)
            dto = {
                'state':'ERROR',
                'context':'None',
                'command': 'None',
                'msg': 'None',
                'options': []
            }
            response = {'requestID': rqstID, 'data': dto}
            self.adm.sendEvent('front_controller', 'ws_response', response)
    




from ...kernel.Agent.Action import Action

class WSResponseAction(Action):
    
    def execute(self, data):    	
        queue = self.agent.requestTable[data['requestID']]
        dto = data['data']
        queue.put(dto)
        


class SpiderData():

    def __init__(self, urls = None):
        self.__urls = None
        self.__user_agent_list = None
        if not urls:
            self.__urls = []
    
    def setUrls(self, urls):
        self.__urls = urls

    def addUrl(self, url):
        self.__urls.append(url)

    def getUrls(self):
        return self.__urls
    
    def setUserAgentList(self, user_agent_list):
        self.__user_agent_list = user_agent_list

    def getUserAgentList(self):
        return self.__user_agent_list
